'use strict';

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

var StringNumeralSystem = require('string-numeral-system');

var jsonpatch = require('fast-json-patch');
var msgpack = require('msgpack-lite');

var PACK_DIFF_KEYS = {
    value: 'v',
    op: 'o',
    path: 'p',
    v: 'value',
    o: 'op',
    p: 'path'
};
var UNPACK_DIFF_KEYS = {
    value: 'value',
    op: 'op',
    path: 'path',
    v: 'v',
    o: 'o',
    p: 'p'
};

function invertObject(obj) {
    return Object.keys(obj).reduce(function (output, key) {
        var value = obj[key];

        output[value] = key;

        return output;
    }, {});
}

var OPERATIONS_KEYS = {
    add: 'a',
    remove: 'r',
    replace: 'p',
    move: 'm',
    copy: 'c',
    test: 't'
};
var INVERTED_OPERATIONS_KEYS = invertObject(OPERATIONS_KEYS);

function BinaryDiffProtocol(compressionMapping, throwOnUnmappedKey) {
    var decompressionMapping = invertObject(compressionMapping);

    function transformDiff(diff, keyMapping, operationsMapping, mapping) {
        var result = [];

        var _iteratorNormalCompletion = true;
        var _didIteratorError = false;
        var _iteratorError = undefined;

        try {
            for (var _iterator = diff[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                var patch = _step.value;

                var transformedDiff = _defineProperty({}, keyMapping.path, '');

                var keys = patch[keyMapping.p].substring(1).split('/');

                var _iteratorNormalCompletion2 = true;
                var _didIteratorError2 = false;
                var _iteratorError2 = undefined;

                try {
                    for (var _iterator2 = keys[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
                        var key = _step2.value;

                        var mappedKey = mapping[key];

                        if (!mappedKey && throwOnUnmappedKey) {
                            throw new Error('Unmapped key: "' + key + '".');
                        }

                        transformedDiff[keyMapping.path] += '/' + (mappedKey || key);
                    }
                } catch (err) {
                    _didIteratorError2 = true;
                    _iteratorError2 = err;
                } finally {
                    try {
                        if (!_iteratorNormalCompletion2 && _iterator2.return) {
                            _iterator2.return();
                        }
                    } finally {
                        if (_didIteratorError2) {
                            throw _iteratorError2;
                        }
                    }
                }

                transformedDiff[keyMapping.value] = patch[keyMapping.v];
                transformedDiff[keyMapping.op] = operationsMapping[patch[keyMapping.o]];

                result.push(transformedDiff);
            }
        } catch (err) {
            _didIteratorError = true;
            _iteratorError = err;
        } finally {
            try {
                if (!_iteratorNormalCompletion && _iterator.return) {
                    _iterator.return();
                }
            } finally {
                if (_didIteratorError) {
                    throw _iteratorError;
                }
            }
        }

        return result;
    }

    function compress(diff) {
        return transformDiff(diff, PACK_DIFF_KEYS, OPERATIONS_KEYS, compressionMapping);
    }

    function decompress(diff) {
        return transformDiff(diff, UNPACK_DIFF_KEYS, INVERTED_OPERATIONS_KEYS, decompressionMapping);
    }

    function pack(data, previousData, inspect) {
        var diff = jsonpatch.compare(previousData || {}, data);
        var compressedDiff = compress(diff);
        var encodedDiff = msgpack.encode(compressedDiff);

        if (typeof inspect === 'function') {
            inspect(data, diff, compressedDiff, encodedDiff);
        }

        return encodedDiff;
    }

    function unpack(binaryData, previousData, inspect) {
        var encodedDiff = new Uint8Array(binaryData);
        var compressedDiff = msgpack.decode(encodedDiff);
        var diff = decompress(compressedDiff);
        var data = Object.assign({}, previousData);

        jsonpatch.apply(data, diff);

        if (typeof inspect === 'function') {
            inspect(data, diff, compressedDiff, encodedDiff);
        }

        return data;
    }

    return {
        pack: pack,
        unpack: unpack
    };
}

var CHARS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=';

/**
 * Generate a mapping for the compression of pack and unpacking. An existing mapping
 * can be provided to expand on, if it was generated using the same character set.
 *
 * @param {array} originalKeys - Keys to generate a mapping for.
 * @param {object} [existingMapping] - An existing mapping to expand on.
 * @param {string} chars - A custom character set to use for the generated keys.
 *
 * @returns {object} The generated mapping.
 */
function generateMapping(originalKeys, existingMapping, chars) {
    chars = chars || CHARS;

    var sns = StringNumeralSystem(chars);
    var mapping = Object.assign({}, existingMapping);

    var last = chars[0];

    if (existingMapping) {
        var _iteratorNormalCompletion3 = true;
        var _didIteratorError3 = false;
        var _iteratorError3 = undefined;

        try {
            for (var _iterator3 = Object.keys(existingMapping)[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
                var existingKey = _step3.value;

                var value = existingMapping[existingKey];

                if (sns.max(value, last) === value) {
                    last = sns.increment(value);
                }
            }
        } catch (err) {
            _didIteratorError3 = true;
            _iteratorError3 = err;
        } finally {
            try {
                if (!_iteratorNormalCompletion3 && _iterator3.return) {
                    _iterator3.return();
                }
            } finally {
                if (_didIteratorError3) {
                    throw _iteratorError3;
                }
            }
        }
    }

    var _iteratorNormalCompletion4 = true;
    var _didIteratorError4 = false;
    var _iteratorError4 = undefined;

    try {
        for (var _iterator4 = originalKeys[Symbol.iterator](), _step4; !(_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done); _iteratorNormalCompletion4 = true) {
            var originalKey = _step4.value;

            if (!mapping[originalKey]) {
                mapping[originalKey] = last;

                last = sns.increment(last);
            }
        }
    } catch (err) {
        _didIteratorError4 = true;
        _iteratorError4 = err;
    } finally {
        try {
            if (!_iteratorNormalCompletion4 && _iterator4.return) {
                _iterator4.return();
            }
        } finally {
            if (_didIteratorError4) {
                throw _iteratorError4;
            }
        }
    }

    return mapping;
}

module.exports = { create: BinaryDiffProtocol, generateMapping: generateMapping };